
/*
 *  shm_msgclient.c
 *  
 *  Illustrates memory mapping and persistency, with POSIX objects.
 *  This process reads and displays a message left it in "memory segment
 *  image", a file been mapped from a memory segment.
 *
 *
 *  Note from nieminen@jyu.fi: This version of Mij's public domain
 *  example is adapted for our spring 2015 course on OS fundamentals.
 *  Please see the original site for more information!
 *
 *
 *  Created by Mij <mij@bitchx.it> on 27/08/05.
 *  Original source file available at http://mij.oltrelinux.com/devel/unixprg/
 *
 */

#define _XOPEN_SOURCE 700 /* Single UNIX Specification, Version x */
/* (See POSIX 2008.1 rationale chapter B.2.2 ) */

#include <stdio.h>
/* exit() etc */
#include <unistd.h>
/* shm_* stuff, and mmap() */
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
/* for random() stuff */
#include <stdlib.h>
#include <time.h>


/* Posix IPC object name [system dependant] - see
http://mij.oltrelinux.com/devel/unixprg/index2.html#ipc__posix_objects */
#define SHMOBJ_PATH         "/foo1423"      
/* maximum length of the content of the message */
#define MAX_MSG_LENGTH      50
/* how many types of messages we recognize (fantasy) */
#define TYPES               8

/* message structure for messages in the shared segment */
struct msg_s {
    int type;
    char content[MAX_MSG_LENGTH];
};

/* Pause to illustrate the race more often:*/
void pikkuinen_tauko(int nsec) {
    struct timespec tim;
    tim.tv_sec = 0;
    tim.tv_nsec = nsec;
    nanosleep(&tim,NULL);
}

int main(int argc, char *argv[]) {
    int shmfd;
    int shared_seg_size = (1 * sizeof(struct msg_s));   /* want shared segment capable of storing 1 message */
    struct msg_s *shared_msg;      /* the shared segment, and head of the messages list */
    

    /* creating the shared memory object    --  shm_open()  */
    shmfd = shm_open(SHMOBJ_PATH, O_RDWR, S_IRWXU | S_IRWXG);
    if (shmfd < 0) {
        perror("In shm_open()");
        exit(1);
    }
    printf("Created shared memory object %s\n", SHMOBJ_PATH);
    
    /* requesting the shared segment    --  mmap() */    
    shared_msg = (struct msg_s *)mmap(NULL, shared_seg_size, PROT_READ | PROT_WRITE, MAP_SHARED, shmfd, 0);
    if (shared_msg == NULL) {
        perror("In mmap()");
        exit(1);
    }
    printf("Shared memory segment allocated correctly (%d bytes).\n", shared_seg_size);


    printf("Message type is %d, content is: %s\n", shared_msg->type, shared_msg->content);
    /* nieminen 2015: Alter the message, in place (with a bad race
     * with multiple concurrent clients!
     * 
     * Exercise for those interested: Implement a mutex around this
     * part. Hint: Since the mutex is between processes, you'll need
     * to learn a bit more about POSIX mutexes than is offered on our
     * intro course. For example, the Stackoverflow has nice questions
     * and answers about this):
     */
    char *alt = shared_msg->content;
    if (*alt != '\0') {
        char rot = alt[0];
        for (int i=0;i<MAX_MSG_LENGTH-1;i++) {
            if (alt[i+1]=='\0'){
                alt[i] = rot;
                break;
            }
            alt[i]=alt[i+1];
            pikkuinen_tauko(100000+(rand()%150000));
        }
    }
    
    return 0;
}

